"""Test of metric functionality."""
import numpy as np
import pandas as pd

from .targets import create_classification_target, create_regression_target
from .errors import TestError
from ..base.metric import Metric
from ..base.types import Metatype, TARGET_METATYPES

def test_metric(metric, n_outputs = 3):
    """
    Function for testing metric functionality.

    Args:
        metric: ClassificationMetric or RegressionMetric.
        n_outputs: dimensionality of targets

    Returns:
        None.
    """
    test_metatypes = [metatype for metatype in TARGET_METATYPES if metric.supports_metatype(metatype)]
    if not test_metatypes:
        raise ValueError('no supported metatypes')

    for metatype in test_metatypes:
        if metric.supports_metatype(metatype):
            if metatype == Metatype.NUMERICAL:
                _test_regression_metric(metric, n_cols = n_outputs)
            else:
                if metatype == Metatype.BINARY:
                    _test_classification_metric(metric, ['good', 'bad'], n_cols = n_outputs)
                else:
                    _test_classification_metric(metric, ['dog', 'cat', 'mouse', 'elephant'], n_cols = n_outputs)


def _test_classification_metric(metric, categories, n_rows = 1000, n_cols = 3):
    index = range(n_rows)

    n_categories = len(categories)

    if len(set(categories)) != n_categories:
        raise TestError('categories must be unique')

    y = create_classification_target(index, n_cols, categories, 0)
    if metric._prediction_method == 'predict_proba':
        if n_cols == 1:
            p = np.random.rand(n_rows, n_categories)
            p /= p.sum(axis=1).reshape(-1,1)
            pred = pd.DataFrame(p, index = index, columns = categories)
        else:
            pred = []
            for i in range(n_cols):
                p = np.random.rand(n_rows, n_categories)
                p /= p.sum(axis=1).reshape(-1,1)
                pred.append(pd.DataFrame(p, index = index, columns = categories))
    elif metric._prediction_method == 'decision_function':
        if n_cols == 1:
            pred = pd.DataFrame(np.random.randn(n_rows, n_categories), index = index, columns = categories)
        else:
            pred = [pd.DataFrame(np.random.randn(n_rows, n_categories), index = index, columns = categories) for i in range(n_cols)]
    else:
        pred = create_classification_target(index, n_cols, categories, 1)
    score = metric.score(y, pred)
    print(f'TESTING DONE: {metric}={score}')

def _test_regression_metric(metric, n_rows = 1000, n_cols = 3):
    index = range(n_rows)
    y = create_regression_target(index, n_cols, 0)
    pred = create_regression_target(index, n_cols, 1)
    score = metric.score(y, pred)
    print(f'TESTING DONE: {metric}={score}')


"""Predictor base class."""
from abc import ABC, abstractmethod

class Predictor(ABC):
    """Base class for all predictors."""
    @classmethod
    @abstractmethod
    def load(cls, filename):
        """
        Loads a model.

        Args:
            filename: string.

        Returns:
            Model.
        """
        pass

    @abstractmethod
    def save(self, filename):
        """
        Saves model to file.

        Args:
            filename: string.

        Returns:
            None.
        """
        pass


import numpy as np
import pandas as pd

def create_classification_target(index, n_cols, categories, seed = 0):
    def transform(x):
        if pd.isnull(x):
            return x
        return categories[x]

    rng = np.random.default_rng(seed = seed)
    return pd.DataFrame(rng.integers(0, len(categories), (len(index), n_cols)), index = index, columns = [f'T{i}' for i in range(n_cols)]).applymap(transform)


def create_regression_target(index, n_cols, seed = 0):
    rng = np.random.default_rng(seed = seed)
    return pd.DataFrame(rng.normal(size = (len(index), n_cols)) ** 2 + np.arange(1, n_cols+1), index = index, columns = [f'T{i}' for i in range(n_cols)])



"""Testing of tabular algos."""
import uuid
import time
import pickle
import os
import numpy as np
import pandas as pd
import datetime

from .errors import TestError
from .targets import create_classification_target, create_regression_target
from ..base.tabular import TabularClassificationAutoML, TabularRegressionAutoML, TabularClassifier, TabularRegressor, is_input_metatype, extract_statics
from ..base.types import Metatype, is_classification_metatype, is_delimiter_metatype, get_delimiter
from ..base.errors import ValidationError
from ..metrics.metrics import get_metric

def create_tabular_inputs(input_metatypes, n_rows = 1000, n_cols = 2, seed = 0):
    X_metatypes = []
    data = []
    index = range(n_rows)

    for metatype in input_metatypes:
        X_metatypes += [metatype for _ in range(n_cols)]

        rng = np.random.default_rng(seed = seed)

        if metatype == Metatype.NUMERICAL:
            x = create_regression_target(index, n_cols, seed = seed)
        elif metatype == Metatype.CATEGORICAL:
            x = create_classification_target(index, n_cols, ['red', 'green', 'blue', 'yellow'], seed = seed)
        elif metatype == Metatype.BINARY:
            x = create_classification_target(index, n_cols, ['bad', 'good'], seed = seed)
        elif metatype == Metatype.DATETIME:
            timestamps = pd.DataFrame(datetime.datetime(2022, 2, 22, 0, 0).timestamp() + rng.normal(0, 1e6, (n_rows, n_cols)))
            x = timestamps.applymap(lambda x: datetime.datetime.fromtimestamp(x).isoformat())
        elif metatype == Metatype.TIMESTAMP:
            x = pd.DataFrame(datetime.datetime(2022, 2, 22, 0, 0).timestamp() + rng.normal(0, 1e6, (n_rows, n_cols)))
        elif is_delimiter_metatype(metatype, strict = True):
            delim = get_delimiter(metatype)
            categories = ['dog','cat','dolphin','lion']
            idx = pd.DataFrame(rng.integers(1,len(categories),size=(1000, n_cols)))
            x = idx.applymap(lambda x: delim.join(categories[:x]))
        else:
            raise TestError('invalid metatype')

        data.append(x)

    X = pd.concat(data, axis = 1)
    X.columns = [f'X{i}' for i in range(X.shape[1])]
    return (X.astype(str), X_metatypes)

def load_tabular_data(algo, nan_fraction = 0.1):
    X, X_metatypes = create_tabular_inputs(algo.input_metatypes())

    rng = np.random.default_rng(seed = 0)
    X.mask(rng.uniform(size = X.shape) < nan_fraction, inplace=True)

    is_multioutput = algo.is_multioutput()
    if not isinstance(is_multioutput, bool):
        raise TestError('is_multioutput should return bool')

    n_cols = 3 if is_multioutput else 1

    if issubclass(algo, TabularClassificationAutoML):
        target_metatype = algo.target_metatype()
        categories = ['bad', 'good'] if target_metatype == Metatype.BINARY else ['red', 'green', 'blue', 'yellow']
        y = create_classification_target(X.index, n_cols, categories)
        y_metatypes = [target_metatype for x in range(n_cols)]
    elif issubclass(algo, TabularRegressionAutoML):
        y = create_regression_target(X.index, n_cols)
        y_metatypes = [Metatype.NUMERICAL for x in range(n_cols)]
    else:
        raise TypeError('invalid algo class')

    return (X, y, X_metatypes, y_metatypes)

def test_tabular_algo(algo, output_dir, timeout, nan_fraction):
    """
    Tests tabular algo.
    
    Args:
        algo: TabularClassificationAutoML or TabularRegressionAutoML.
        output_dir: path for temp files.   
        timeout: training time in seconds.
        nan_fraction: fraction of missing observations in inputs.

    Raises:
        TypeError, TestError
    """
    if not (issubclass(algo, TabularClassificationAutoML) or issubclass(algo, TabularRegressionAutoML)):
        raise TypeError('invalid algo class')

    statics = extract_statics(algo)

    if issubclass(algo, TabularClassificationAutoML):
        metric = get_metric('accuracy')
    else:
        metric = get_metric('mean_squared_error')

    X, y, X_metatypes, y_metatypes = load_tabular_data(algo, nan_fraction)

    model = algo()

    t_start = time.time()
    model.fit(timeout, metric, X, y, X_metatypes, y_metatypes, X_unlabelled = X)
    fit_seconds = time.time() - t_start

    if fit_seconds > timeout:
        raise TestError(f'timeout exceeded by {fit_seconds - timeout} seconds')

    print(f'Elapsed time: {fit_seconds}, timeout: {timeout}')

    test_tabular_predictor(model, output_dir, X, y)

def test_tabular_predictor(predictor, output_dir, X, y):
    """
    Tests tabular predictor.
    
    Args:
        predictor: TabularClassifier or TabularRegressor.
        output_dir: path for temp files.   
        X: DataFrame.
        y: DataFrame.

    Raises:
        TypeError, TestError
    """
    if not (isinstance(predictor, TabularClassifier) or isinstance(predictor, TabularRegressor)):
        raise TypeError('invalid predictor')

    if not isinstance(X, pd.DataFrame):
        raise TypeError('X must be a DataFrame')

    if not isinstance(y, pd.DataFrame):
        raise TypeError('y must be a DataFrame')

    pred = predictor.predict(X)
    if not isinstance(pred, pd.DataFrame):
        raise TestError('prediction must be DataFrame')

    if list(pred.columns) != list(y.columns):
        raise TestError('prediction columns must equal y columns')

    if list(pred.index) != list(X.index):
        raise TestError('prediction index must equal X index')

    if isinstance(predictor, TabularClassifier):
        prediction_methods = [x.name.lower() for x in predictor.prediction_methods()]
        for prediction_method in prediction_methods:
            if prediction_method != 'predict':
                pred_prediction_method = getattr(predictor, prediction_method)(X)
                if y.shape[1] == 1:
                    pred_prediction_method = [pred_prediction_method]

                for ppm, col in zip(pred_prediction_method, y):
                    _check_prediction(prediction_method, ppm, X, y[col], pred[col])

    if output_dir is not None:
        filename = os.path.join(output_dir, uuid.uuid4().hex + '.mdl')
        predictor.save(filename)

        model_loaded = predictor.load(filename)
        pred_loaded = model_loaded.predict(X)

        if not pred_loaded.equals(pred):
            raise TestError('loaded model predictions are off')

        os.remove(filename)

    print(f'TESTING DONE: {type(predictor).__name__}')


def _check_prediction(prediction_method, ppm, X, y, pred):
    if not isinstance(ppm, pd.DataFrame):
        raise TestError(f'{prediction_method} must return a DataFrame for univariate and list thereof for multivariate problems')

    if list(ppm.index) != list(X.index):
        raise TestError('prediction index must equal X index')

    if prediction_method == 'predict_proba':
        if (ppm < 0).any(axis = None):
            raise TestError(f'{prediction_method} probabilities cannot be negative')

        if (ppm > 1).any(axis = None):
            raise TestError(f'{prediction_method} probabilities cannot be greater than 1')

        if ((ppm.sum(axis=1) - 1).abs() > 1e-6).any():
            raise TestError(f'{prediction_method} probabilities must sum to 1 for each row')

    if prediction_method in ('predict_proba', 'decision_function'):
        idx = ppm.to_numpy().argmax(axis = 1)
        if ppm.columns[idx].tolist() != pred.tolist():
            raise TestError(f'argmax of {prediction_method} output should equal predicted class')

    y_unique = set(y)

    if ppm.shape[1] != len(y_unique) or set(ppm.columns) != y_unique:
        raise TestError(f'{prediction_method} must have a column for each class')





"""Metric base class."""
from abc import ABC, abstractmethod
from .types import Metatype, is_target_metatype

class Metric(ABC):
    """Base class for all Metrics."""
    @abstractmethod 
    def __init__(self, name, metatypes, prediction_method, greater_is_better):
        """
        Metric base class.

        Args:
            name: string representation.
            metatypes: Metatype or set thereof.
            prediction_method: predict, predict_proba, decision_function.
            greater_is_better: whether Metric is a score function (higher is better), or loss function (lower is better).
        """
        if not isinstance(name, str):
            raise TypeError('name must be a string')
        self._name = name

        if isinstance(metatypes, Metatype):
            metatypes = {metatypes}

        if not isinstance(metatypes, set):
            raise TypeError('metatypes must be a set')

        for metatype in metatypes:
            if not is_target_metatype(metatype, strict = True):
                raise ValueError('metatypes must be a set of Metatypes')

        if prediction_method not in ('predict', 'predict_proba', 'decision_function'):
            raise ValueError('invalid prediction_method')

        if Metatype.NUMERICAL in metatypes and prediction_method != 'predict':
            raise ValueError('prediction_method must be predict for numreical-metatype')

        self._metatypes = metatypes
        self._prediction_method = prediction_method
        self._sign = -1 if greater_is_better else 1

    @abstractmethod
    def score(self, y, pred):
        """
        Invokes underlying score function.

        Args:
            y: targets.
            pred: model predictions.

        Returns:
            float.
        """
        pass

    def __str__(self):
        """Name of Metric."""
        return self._name

    def to_dict(self):
        metatypes = [metatype.name.lower() for metatype in self._metatypes]
        return dict(name = self._name, prediction_method = self._prediction_method, sign = self._sign, metatypes = metatypes)

    def loss(self, y, pred):
        """
        Loss based on underlying score function where lower is better.

        Args:
            y: targets.
            pred: model predictions.

        Returns:
            float.
        """
        return self.score_to_loss(self.score(y, pred))

    def score_to_loss(self, score):
        """
        Converts score to loss.

        Args:
            score: float.

        Returns:
            float.
        """
        return self._sign * score

    def prediction_score(self, model, X, y):
        """
        Combined prediction and scoring.

        Args:
            model: estimator with prediction method.
            X: inputs.
            y: targets.

        Returns:
            float.
        """
        return self.score(y, self.prediction(model, X))

    def prediction_loss(self, model, X, y):
        """
        Combined prediction and loss.

        Args:
            model: estimator with prediction method.
            X: inputs.
            y: targets.

        Returns:
            float.
        """
        return self.loss(y, self.prediction(model, X))

    def prediction(self, model, *args, **kwargs):
        """
        Returns prediction of model.

        Args:
            model: implements prediction method.
            *args: arguments to prediction method
            **kwargs: keyword arguments to prediction method

        Returns:
            model prediction.
        """
        return getattr(model, self._prediction_method)(*args, **kwargs)

    def supports_metatype(self, metatype):
        """
        Checks if metric supports a given metatype.

        Args:
            metatype: Metatype

        Returns:
            bool.
        """
        return metatype in self._metatypes



"""AlgoSquare Dataset API."""
import io
import time
import numpy as np
import pandas as pd
import requests
import json

from ..base.tabular import is_input_metatype
from ..base.types import Metatype, is_target_metatype, is_classification_metatype, string_to_metatype
from ..metrics import get_metric

from .api import ApiObject, api_get, api_post, api_put
from .task import Task
from .common import upload_dataframe

def get_datasets():
    """fetches all datasets."""
    return [Dataset.load(x) for x in api_get('api/datasets')]

class Dataset(ApiObject):
    """Base class for all datasets."""
    def __init__(self, entry):
        if type(self) == Dataset:
            raise RuntimeError('base class should not be instantiated')
        super().__init__(entry)

    @staticmethod
    def get(dataset_id):
        """
        Gets specific dataset.
        
        Args:
            dataset_id: string.

        Returns:
            Dataset subclass.
        """
        return Dataset.load(api_get(f'api/datasets/{dataset_id}'))

    @staticmethod
    def load(entry):
        if entry['data_format'] == 'tabular':
            return TabularData(entry)
        else:
            raise ValueError('unknown data_format')

    def refresh(self):
        """reloads dataset."""
        self.update(api_get(f'api/datasets/{self.dataset_id}'))

    def get_tasks(self):
        """
        Gets tasks for dataset.
        
        Returns:
            List of Tasks.
        """
        return [Task(x) for x in api_get(f'api/datasets/{self.dataset_id}/tasks')]

class TabularData(Dataset):
    """Tabular data class."""
    @classmethod
    def from_csv(cls, input_file, target_file, title = None, input_metatypes = None, target_metatypes = None):
        """
        Creates tabular data from csv files.
        
        Args:
            input_file: path to input data.
            target_file: path to target data.
            title: title of dataset.
            input_metatypes: list of Metatype or path to csv file.
            target_metatypes: list of Metatype or path to csv file.

        Returns:
            TabularData.
        """
        input_df = pd.read_csv(input_file, dtype=str)
        target_df = pd.read_csv(target_file, dtype=str)
        return cls.from_df(title, input_df, target_df, input_metatypes, target_metatypes)

    @classmethod
    def from_df(cls, input_df, target_df, title = None, input_metatypes = None, target_metatypes = None, ):
        """
        Creates tabular data from DataFrames.
        
        Args:
            input_df: DataFrame with input data.
            target_df: DataFrame with target data.
            title: title of dataset.
            input_metatypes: list of Metatype or path to csv file.
            target_metatypes: list of Metatype or path to csv file.

        Returns:
            TabularData.

        Raises:
            TypeError, ValueError
        """
        if not isinstance(title, str):
            raise TypeError('title must be a string')
        if len(title) > 1024:
            raise ValueError('title must be maximum 1024 characters')

        if not isinstance(input_df, pd.DataFrame):
            raise TypeError('input_df must be a DataFrame')
        if not isinstance(target_df, pd.DataFrame):
            raise TypeError('target_df must be a DataFrame')

        metatypes = []

        if input_metatypes is not None:
            prepared_metatypes = _prepare_metatypes(input_df, input_metatypes)
            for metatype in prepared_metatypes.values():
                if not is_input_metatype(metatype):
                    raise ValueError(f'invalid input metatype: {metatype}')
            metatypes.append(dict(namespace = 'inputs', metadata = [dict(name = name, metatype = prepared_metatypes.get(name, Metatype.VOID).name.lower()) for name in input_df]))

        if target_metatypes is not None:
            prepared_metatypes = _prepare_metatypes(target_df, target_metatypes)
            for metatype in prepared_metatypes.values():
                if not is_target_metatype(metatype):
                    raise ValueError(f'invalid target metatype: {metatype}')
            metatypes.append(dict(namespace = 'targets', metadata = [dict(name = name, metatype = prepared_metatypes.get(name, Metatype.VOID).name.lower()) for name in target_df]))

        files = [dict(namespace = namespace, key = upload_dataframe(namespace, df)) for namespace, df in [('inputs', input_df), ('targets', target_df)]]
        
        settings = dict(data_format = 'tabular')
        if title:
            if not isinstance(title, str) and len(title) > 128:
                raise ValueError('invalid title')
            settings['title'] = title
        
        entry = api_post('api/datasets', json=dict(settings = settings, files = files, metatypes = metatypes))
        return cls(entry)

    def create_task(self, title = None, metric_name = None, pos_label = None, holdout = 25, input_columns = None, target_columns = None):
        """
        Creates a task for dataset.
        
        Args:
            title: title of task.
            metric_name: name used for get_metric.
            pos_label: positive label for binary metrics.
            holdout: size of test set as a proportion of dataset.
            input_columns: list of columns in inputs.
            target_columns: list of columns in targets.

        Returns:
            Task.

        Raises:
            RuntimeError, ValueError
        """
        if self.status != 'ok':
            raise RuntimeError('status must be ok')

        holdout = int(holdout)
        if holdout < 5 or holdout > 30:
            raise ValueError('holdout outside of range 5-30')

        if input_columns is None:
            input_columns = [x['name'] for x in self.metadata['inputs']]

        if target_columns is None:
            target_columns = [self.metadata['targets'][0]['name']]

        if len(target_columns) != 1:
            raise ValueError('only one target column is currently allowed')

        features = dict(inputs = [x for x in self.metadata['inputs'] if x['name'] in input_columns], targets = [x for x in self.metadata['targets'] if x['name'] in target_columns])

        if len(features['inputs']) == 0:
            raise ValueError('no X features selected')

        if len(features['targets']) != 1:
            raise ValueError('only one y feature must be selected')

        target_metatype = string_to_metatype(features['targets'][0]['metatype'])
        if not is_target_metatype(target_metatype):
            raise ValueError('invalid target_metatype')

        if metric_name is None:
            metric_name = 'accuracy' if is_classification_metatype(target_metatype) else 'mean_squared_error'
        else:
            metric = get_metric(metric_name)
            if metric['metatypes'] == ['binary'] and pos_label is None:
                raise ValueError('pos_label is missing')

        settings = dict(metric = metric_name, holdout = holdout)
        if title:
            if not isinstance(title, str) and len(title) > 128:
                raise ValueError('invalid title')
            settings['title'] = title

        if pos_label is not None:
            if target_metatype != Metatype.BINARY:
                raise ValueError('pos_label only expected for binary metatypes')

            categories = features['targets'][0]['stats']['categories']
            if pos_label not in categories:
                raise RuntimeError(f'{pos_label} not in {categories}')
            settings['pos_label'] = pos_label

        payload = dict()
        payload['settings'] = settings
        payload['selections'] = [dict(namespace = namespace, features = [x['feature_id'] for x in v]) for namespace, v in features.items()]

        return Task(api_post(f'api/datasets/{self.dataset_id}/tasks', json=payload))

def _prepare_metatypes(data_df, metatypes):
    if isinstance(metatypes, str):
        mt_df = pd.read_csv(metatypes, dtype = str, index_col = 0)
        if mt_df.shape[1] != 1:
            raise ValueError(f'{metatypes} must have exactly two columns')
        mt_dict = mt_df.squeeze().to_dict()
        if set(mt_dict.keys()).difference(data_df):
            raise ValueError(f'{metatypes} have invalid names')

        return {k : string_to_metatype(v) for k, v in mt_dict.items()}
    else:
        return dict(zip(data_df, metatypes))

def _sizeof_fmt(num, suffix="B"):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi']:
        if abs(num) < 1024.0 or unit == 'Pi':
            break
        num /= 1024.0
    return f'{num:3.1f} {unit}{suffix}'


"""Tabular algos and auxiliary functions."""

from abc import ABC, abstractmethod
from .algo import Algo
from .predictor import Predictor
from .types import Metatype, PredictionMethod, is_classification_metatype, is_prediction_method
from .errors import ValidationError

INPUT_METATYPES = {Metatype.NUMERICAL, Metatype.CATEGORICAL, Metatype.BINARY,
            Metatype.DATETIME, Metatype.TIMESTAMP, Metatype.DELIM_PIPE, Metatype.DELIM_SEMICOLON}

class TabularClassifier(Predictor, ABC):
    """
    Base class for tabular classification model.

    Note:
        'predict_proba' and 'decision_function' are optional, but should be implemented if possible.
    """
    @staticmethod
    @abstractmethod
    def prediction_methods():
        """
        Describes the prediction_methods that the classifier supports.
        
        Returns: 
            Set of PredictionMethods.
        """
        pass

    @abstractmethod
    def predict(self, X):
        """
        Makes prediction for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of y.
        """
        pass

    def predict_proba(self, X):
        """
        Outputs class probabilities for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of classes in y, or list thereof for multivariate problems.

        Raises:
            NotImplementedError.
        """
        raise NotImplementedError

    def decision_function(self, X):
        """
        Outputs decision function for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of classes in y, or list thereof for multivariate problems.
        
        Raises:
            NotImplementedError.
        """
        return self.predict_proba(X)

class TabularClassificationAutoML(Algo, TabularClassifier, ABC):
    """Base class for tabular classification learning algorithm."""
    @abstractmethod
    def fit(self, timeout, metric, X, y, X_metatypes, y_metatypes, X_unlabelled = None):
        """
        Fits model on data.

        Args:
            timeout: time in seconds to fit the model.
            metric: Metric.
            X: DataFrame with shape (n_samples, n_features) and dtype=str.
            y: DataFrame with shape (n_samples, n_outputs) and dtype=str.
            X_metatypes: list of n_features Metatypes.
            y_metatypes: list of n_outputs Metatypes.
            X_unlabelled: DataFrame with shape (n_samples_unlabelled, n_features).
        """
        pass

    @staticmethod
    @abstractmethod
    def input_metatypes():
        """
        Describes the X_metatypes that the algo supports.
        
        Returns: 
            Set of Metatypes.
        """
        pass

    @staticmethod
    @abstractmethod
    def target_metatype():
        """
        Describes the y_metatype that the algo supports.
        
        Returns: 
            Metatype.
        """
        pass

    @staticmethod
    @abstractmethod
    def is_multioutput():
        """
        Returns true if the algo supports mulitple outputs.
        
        Returns: 
            Boolean.
        """
        pass

class TabularRegressor(Predictor, ABC):
    """Base class for tabular regression model."""
    @abstractmethod
    def predict(self, X):
        """
        Makes prediction for given data.
        
        Args:
            X: DataFrame with shape (n_samples, n_features).

        Returns:
            DataFrame with index of X and columns of y.
        """
        pass

class TabularRegressionAutoML(Algo, TabularRegressor, ABC):
    """Base class for tabular regression learning algorithm."""
    @abstractmethod
    def fit(self, timeout, metric, X, y, X_metatypes, y_metatypes, X_unlabelled = None):
        """
        Fits model on data.

        Args:
            timeout: time in seconds to fit the model.
            metric: Metric.
            X: DataFrame with shape (n_samples, n_features) and dtype=str.
            y: DataFrame with shape (n_samples, n_outputs) and dtype=str.
            X_metatypes: list of n_features Metatypes.
            y_metatypes: list of n_outputs Metatypes.
            X_unlabelled: DataFrame with shape (n_samples_unlabelled, n_features).
        """
        pass

    @staticmethod
    @abstractmethod
    def input_metatypes():
        """
        Describes the X_metatypes that the algo supports.
        
        Returns: 
            Set of Metatypes.
        """
        pass

    @staticmethod
    @abstractmethod
    def is_multioutput():
        """
        Returns true if the algo supports mulitple outputs.
        
        Returns: 
            Boolean.
        """
        pass

def is_input_metatype(metatype, strict = False):
    """Checks if metatype is valid input metatype.

    Args:
        metatype: Metatype or int.
        strict: checks if metatype is Metatype.

    Returns: 
        bool.
    """
    if strict and not isinstance(metatype, Metatype):
        return False

    try:
        metatype = Metatype(metatype)
        return metatype in INPUT_METATYPES
    except ValueError:
        return False

def extract_statics(obj):
    statics = dict()

    if issubclass(obj, TabularClassificationAutoML) or issubclass(obj, TabularRegressionAutoML):
        input_metatypes = obj.input_metatypes()
        if not isinstance(input_metatypes, set) or not input_metatypes:
            raise ValidationError('input_metatypes must be a non-empty set')

        for metatype in input_metatypes:
            if not is_input_metatype(metatype, True):
                raise ValidationError('invalid input metatype')

        statics['input_metatypes'] = sorted(list(input_metatypes), key=lambda x: x.name)

        is_multioutput = obj.is_multioutput()
        if not isinstance(is_multioutput, bool):
            raise ValidationError('is_multioutput should return bool')

        statics['is_multioutput'] = is_multioutput

        if issubclass(obj, TabularClassificationAutoML):
            target_metatype = obj.target_metatype()
            if not is_classification_metatype(target_metatype):
                raise ValidationError('invalid classification metatype')

            statics['target_metatype'] = target_metatype


    if issubclass(obj, TabularClassifier):
        prediction_methods = obj.prediction_methods()
        if not isinstance(prediction_methods, set) or not prediction_methods:
            raise ValidationError('prediction_methods must be a non-empty set')

        for prediction_method in prediction_methods:
            if not is_prediction_method(prediction_method, True):
                raise ValidationError('invalid prediction method')

        if PredictionMethod.PREDICT not in prediction_methods:
            raise ValidationError('predict missing from prediction method')

        if PredictionMethod.PREDICT_PROBA in prediction_methods and PredictionMethod.DECISION_FUNCTION not in prediction_methods:
            raise ValidationError('if predict_proba is in prediction_methods, then decision_function should also be')

    else:
        prediction_methods = { PredictionMethod.PREDICT }

    statics['prediction_methods'] = sorted([prediction_method.name.lower() for prediction_method in prediction_methods])

    return statics

    


"""AlgoSquare Model API."""
import io
import json
import pandas as pd

from .api import ApiObject, api_post, api_get, api_delete
from .common import upload_dataframe

def get_deployed():
    """
    Gets all deployed models.
    
    Returns:
        List of Models.
    """
    return [Model.load(x) for x in api_get('api/models/deployments')]

class Model(ApiObject):
    """Base class for all models."""
    def __init__(self, entry):
        if type(self) == Model:
            raise RuntimeError('base class should not be instantiated')
        super().__init__(entry)

    @staticmethod
    def get(model_id):
        """
        Gets specific model.
        
        Args:
            model_id: string.

        Returns:
            Model.
        """
        return Model.load(api_get(f'api/models/{model_id}'))

    @staticmethod
    def load(entry):
        base_class = entry['base_class']
        if base_class in ('TabularClassifier', 'TabularClassificationAutoML'):
            return TabularClassifier(entry)
        elif base_class in ('TabularRegressor', 'TabularRegressionAutoML'):
            return TabularRegressor(entry)
        else:
            raise ValueError('unknown base_class')

    def refresh(self):
        """reloads model."""
        self.update(api_get(f'api/models/{self.model_id}'))

    def deploy(self):
        """
        Deploys model.
        
        Raises:
            RuntimeError.
        """
        if self.status != 'stopped':
            raise RuntimeError('status must be stopped')

        self.update(api_post(f'api/models/{self.model_id}/deployment'))

    def discharge(self):
        """discharges model."""
        if self.status != 'running':
            raise RuntimeError('status must be running')
        self.update(api_delete(f'api/models/{self.model_id}/deployment'))

    def prediction_update(self, prediction_id, actual):
        """
        Updating model.
        
        Args:
            prediction_id: output from model.predict.
            actual: observed target.

        Raises:
            RuntimeError.
        """
        if self.status != 'running':
            raise RuntimeError('status must be running')

        return api_put(f'api/models/{self.model_id}/predictions/{prediction_id}', data = json.dumps(actual))

    def _predict(self, data, prediction_method):
        if self.status != 'running':
            raise RuntimeError('status must be running')

        if prediction_method not in self.prediction_methods:
            raise NotImplementedError(f'{prediction_method} not implemented in model')

        return api_post(f'api/models/{self.model_id}/predictions', data = json.dumps(dict(data = data, prediction_method = prediction_method)))

    def _predict_batch(self, data, prediction_method, upload_func):
        if self.status != 'running':
            raise RuntimeError('status must be running')

        if prediction_method not in self.prediction_methods:
            raise NotImplementedError(f'{prediction_method} not implemented in model')

        files = upload_func(data)
        payload = dict(prediction_method = prediction_method, files = files)

        return api_post(f'api/models/{self.model_id}/batch', data = json.dumps(payload))


class TabularClassifier(Model):
    """Tabular classifier class."""
    def predict(self, data):
        """
        Using deployed model to make single prediction.
        
        Args:
            data: dict

        Returns:
            dict.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, dict):
            raise TypeError('data must be a dict')

        return self._predict(data, 'predict')

    def predict_batch(self, data):
        """
        Using deployed model to make batch prediction.
        
        Args:
            data: DataFrame

        Returns:
            DataFrame.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError('data must be a dataframe')

        return _batch_to_df(self._predict_batch(data, 'predict', _upload_tabular_inputs))

    def predict_proba(self, data):
        """
        Using deployed model to output class probabilities.
        
        Args:
            data: dict

        Returns:
            dict.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, dict):
            raise TypeError('data must be a dict')

        return self._predict(data, 'predict_proba')

    def predict_proba_batch(self, data):
        """
        Using deployed model to output class probabilities for batch.
        
        Args:
            data: DataFrame

        Returns:
            DataFrame.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError('data must be a dataframe')

        return _batch_to_df(self._predict_batch(data, 'predict_proba', _upload_tabular_inputs))

    def decision_function(self, data):
        """
        Using deployed model to output decision function.
        
        Args:
            data: dict

        Returns:
            dict.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, dict):
            raise TypeError('data must be a dict')

        return self._predict(data, 'decision_function')

    def decision_function_batch(self, data):
        """
        Using deployed model to output decision function for batch.

        Args:
            data: DataFrame

        Returns:
            DataFrame.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError('data must be a dataframe')

        return _batch_to_df(self._predict_batch(data, 'decision_function', _upload_tabular_inputs))

class TabularRegressor(Model):
    """Tabular regressor class."""
    def predict(self, data):
        """
        Using deployed model to make single prediction.
        
        Args:
            data: dict

        Returns:
            dict.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, dict):
            raise TypeError('data must be a dict')

        return self._predict(data, 'predict')

    def predict_batch(self, data):
        """
        Using deployed model to make batch prediction.
        
        Args:
            data: DataFrame

        Returns:
            DataFrame.

        Raises:
            NotImplementedError, RuntimeError, TypeError.
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError('data must be a dataframe')

        return _batch_to_df(self._predict_batch(data, 'predict', _upload_tabular_inputs))

def _batch_to_df(response):
    with io.StringIO(response['data']) as buffer:
        return pd.read_csv(buffer, index_col = 0)

def _upload_tabular_inputs(data):
    return [dict(namespace = 'inputs', key = upload_dataframe('inputs', data))]



"""Useful transformers for working with metatypes."""
import datetime
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import LabelEncoder

from ..base.types import Metatype, get_delimiter, is_delimiter_metatype, is_target_metatype, is_classification_metatype

def to_float(df):
    return df.astype(float)

def to_category(df):
    return df.astype('category')

def to_datetime(df):
    return df.apply(pd.to_datetime)

def from_timestamp(df):
    def float_to_timestamp(x):
        if pd.isnull(x):
            return x
        return datetime.datetime.fromtimestamp(x)
    return df.astype(float).applymap(float_to_timestamp)

def to_text(df):
    return df.astype(str).mask(pd.isnull(df))

def from_delim(df, delim):
    def split_delim(x):
        return [y.strip() for y in x.split(delim)]
    return df.astype(str).applymap(split_delim).mask(pd.isnull(df))

def delim_pipe(df):
    return from_delim(df, '|')

def delim_semicolon(df):
    return from_delim(df, ';')

class FunctionTransformer(BaseEstimator, TransformerMixin):
    """
    Transformer that applies a function to X.

    Args:
        func: callable with a single argument that accepts X.

    Note:
        Annonymous functions, i.e lambdas, do not pickle and should be avoided.
    """
    def __init__(self, func):
        self.func = func

    def fit(self, X, y = None):
        return self

    def transform(self, X, y = None):
        """
        Applies function to X.

        Args:
            X: DataFrame or numpy array.
            y: optional.

        Returns:
            func(X).
        """
        return self.func(X)

class CategoricalListTransformer(BaseEstimator, TransformerMixin):
    """
    Converts a list to array with ones where a category is present and zeros if not.

    Args:
        null_as_zeros: missing values transformed to rows of zeros if True, nans otherwise.
    """
    def __init__(self, null_as_zeros = False):
        self.null_as_zeros = null_as_zeros

    def fit(self, X, y = None):
        """
        Fits transformer to X.

        Args:
            X: Series, DataFrame or numpy array.
            y: optional.

        Returns:
            self.
        """
        if isinstance(X, pd.Series):
            X = X.to_frame()

        if isinstance(X, pd.DataFrame):
            self.columns = X.columns
            X = X.to_numpy()

        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.categories = []
        for i in range(X.shape[1]):
            uniques = set()
            for row in X[:,i]:
                if isinstance(row, list):
                    for x in row:
                        uniques.add(x)

            self.categories.append({x : i for i, x in enumerate(uniques)})
        return self

    def transform(self, X):
        """
        Applies function to X.

        Args:
            X: Series, DataFrame or numpy array.
            y: optional.

        Returns:
            numpy array.
        """
        if isinstance(X, pd.Series):
            X = X.to_frame()

        if isinstance(X, pd.DataFrame):
            if list(X.columns) != list(self.columns):
                raise ValueError('when transforming X, columns must be the same as fit') 
            X = X.to_numpy()

        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        transforms = []
        for j, cat in enumerate(self.categories):
            data = np.zeros((X.shape[0], len(cat)))
            for i, row in enumerate(X[:,j]):
                if isinstance(row, list):
                    if row:
                        idx = [cat[x] for x in row if x in cat]
                        data[i,idx] = 1
                elif not self.null_as_zeros:
                    data[i,:] = np.nan
            transforms.append(data)

        return np.concatenate(transforms, axis=1) 

class BinaryTransformer(BaseEstimator, TransformerMixin):
    """
    Converts binary categoricals to bools.

    Args:
        nan_unknowns: replaces unknown category with nan.
    """
    def __init__(self, nan_unknowns = True):
        if not isinstance(nan_unknowns, bool):
            raise TypeError('nan_unkonwns must be bool')
        self.nan_unknowns = nan_unknowns

    def fit(self, X, y = None):
        """
        Fits transformer to X.

        Args:
            X: DataFrame.
            y: optional.

        Returns:
            self.
        """
        self.categories = dict()
        for col, s in X.items():
            cat = s.dropna().unique()
            if len(cat) != 2:
                raise ValueError('non-binary feature: %s' % col)
            self.categories[col] = sorted(list(cat))
        return self

    def transform(self, X):
        """
        Transforms binary categoricals to bool

        Args:
            X: DataFrame.

        Returns:
            DataFrame with object dtype where null values are preserved.
        """
        transforms = []
        for col, s in X.items():
            cat = self.categories[col]
            binary = s == cat[1]
            if self.nan_unknowns:
                binary[~s.isin(cat)] = np.nan

            transforms.append(binary)
        return pd.concat(transforms, axis=1).mask(pd.isnull(X)).astype('O')

class OrdinalEncoder(BaseEstimator, TransformerMixin):
    """Converts categoricals to ordinals and replaces missing and unknowns with nan."""
    def fit(self, X, y = None):
        """
        Fits transformer to X.

        Args:
            X: DataFrame.
            y: optional.

        Returns:
            self.
        """
        self.categories = dict()
        for col, s in X.items():
            self.categories[col] = set(s.dropna().unique())
        return self

    def transform(self, X):
        """
        Ordinal encodes and replaces missing and unknowns with nan.

        Args:
            X: DataFrame.

        Returns:
            DataFrame.
        """
        index = X.index
        columns = X.columns
        X = X.to_numpy()
        
        output = np.empty(X.shape)
        output[:] = np.nan
        
        for i, col in enumerate(columns):
            for j, cat in enumerate(self.categories[col]):
                idx = X[:,i] == cat
                output[idx,i] = j
        
        return pd.DataFrame(output, index = index, columns = columns)

class ColumnTransformer(BaseEstimator, TransformerMixin):
    """ColumnTransformer specifically for DataFrame.

    Args:
        column_transformers: iterable of tuples - see sklearn ColumnTransformer. 

    Note:
        The ColumnTransformer from sklearn interprets integers as positional columns and not column names, and only strings are used to reference DataFrame columns by name. 
        Here, indexes references DataFrame column names in all cases.
    """
    def __init__(self, column_transformers):
        self.column_transformers = column_transformers

    def fit(self, X, y = None):
        """
        Fits transformer to X.

        Args:
            X: DataFrame.
            y: optional.

        Returns:
            self.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError('X should be a DataFrame')

        for name, transformer, columns in self.column_transformers:
            transformer.fit(X[columns], y)

        return self

    def transform(self, X):
        """
        Applies transformers to columns

        Args:
            X: DataFrame.

        Returns:
            numpy array
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError('X should be a DataFrame')

        return np.concatenate([_expand_dim(np.array(transformer.transform(X[columns]))) for name, transformer, columns in self.column_transformers], axis=1)


class MetatypeTransformer(BaseEstimator, TransformerMixin):
    """
    Transformer that converts Metatypes.

    Args:
        X_metatypes: list of Metatypes.
        valid_metatypes: set of Metatypes (optional).
        binary_as_categorical: convert features with two categories to categoricals.

    Note:
        This transformer is intended for input data and it is recommended to use it as the first step in a pipeline. 
    """

    def __init__(self, X_metatypes, valid_metatypes = None, binary_as_categorical = False):
        _check_metatypes(X_metatypes)
        self.X_metatypes = X_metatypes

        if valid_metatypes is not None:
            _check_metatypes(valid_metatypes)
        self.valid_metatypes = valid_metatypes

        self.binary_as_categorical = binary_as_categorical
        self.transformers = []
        self.columns = []

    def _is_valid_metatype(self, metatype):
        return self.valid_metatypes is None or metatype in self.valid_metatypes

    def fit(self, X, y = None):
        """
        Fits transformer to X.

        Args:
            X: DataFrame with same number of columns as X_metatypes.
            y: optional.

        Returns:
            self.
        """
        if not X.shape[1] == len(self.X_metatypes):
            raise ValueError('number of columns must equal length of metatypes')

        def get_transformer(metatype):
            transformers = dict()
            transformers[Metatype.NUMERICAL] = FunctionTransformer(to_float)

            if self.binary_as_categorical:
                transformers[Metatype.BINARY] = FunctionTransformer(to_category)
            else:
                transformers[Metatype.BINARY] = BinaryTransformer()

            transformers[Metatype.CATEGORICAL] = FunctionTransformer(to_category)
            transformers[Metatype.DATETIME] = FunctionTransformer(to_datetime)
            transformers[Metatype.TIMESTAMP] = FunctionTransformer(from_timestamp)
            transformers[Metatype.TEXT] = FunctionTransformer(to_text)
            transformers[Metatype.DELIM_PIPE] = FunctionTransformer(delim_pipe)
            transformers[Metatype.DELIM_SEMICOLON] = FunctionTransformer(delim_semicolon)
            return transformers.get(metatype)

        valid_metatypes = set()
        self.transformers = []
        for metatype, columns in group_metatypes(X, self.X_metatypes).items():
            if self._is_valid_metatype(metatype):
                transformer = get_transformer(metatype)
                if transformer:
                    self.transformers.append((transformer.fit(X[columns]), columns))
                    valid_metatypes.add(metatype)

        self.columns = [col for col, metatype in zip(X, self.X_metatypes) if metatype in valid_metatypes]
        return self

    def transform(self, X):
        """
        Transforms binary categoricals to bool.

        Args:
            X: DataFrame with same number of columns as metatypes.

        Returns:
            DataFrame.

        Note:
            Metatypes 'datetime' and 'timestamp' are both converted to datetime objects, and delimiter metatypes are converted to lists.
        """
        output = pd.concat([transformer.transform(X[columns]) for transformer, columns in self.transformers], axis=1)
        return output[self.columns]

    def transform_metatypes(self):
        """
        Transforms metatypes to output metatypes.

        Returns:
            List of Metatypes.
        """
        def transform_metatype(metatype):
            if metatype == Metatype.TIMESTAMP:
                return Metatype.DATETIME
            if is_delimiter_metatype(metatype):
                return Metatype.LIST
            if metatype == Metatype.BINARY and self.binary_as_categorical:
                return Metatype.CATEGORICAL
            return metatype

        return [transform_metatype(x) for x in self.X_metatypes if self._is_valid_metatype(x)]

class TargetTransformer(BaseEstimator, TransformerMixin):
    """
    Transformer that works for both numerical and categorical targets.

    Args:
        y_metatypes: list of metatype strings ('numerical', 'binary' or 'categorical').

    Note:
        This transformer is intended to be used for translation between underlying model and output of algo. 
    """

    @staticmethod
    def _label_transform(s, encoder):
        return encoder.transform(s)

    @staticmethod
    def _label_inverse_transform(s, encoder):
        return encoder.inverse_transform(s.ravel().astype(int))

    def __init__(self, y_metatypes):
        for metatype in y_metatypes:
            if not is_target_metatype(metatype):
                raise ValueError('invalid target metatype')
        self.y_metatypes = y_metatypes

    def fit(self, y):
        """
        Fits transformer to y.

        Args:
            y: DataFrame with same number of columns as y_metatypes.

        Returns:
            self.
        """
        if not isinstance(y, pd.DataFrame):
            raise TypeError('y must be a dataframe')
        if y.shape[1] != len(self.y_metatypes):
            raise ValueError('wrong number of columns')

        self.label_encoders = {col : (metatype, LabelEncoder()) for col, metatype in zip(y, self.y_metatypes) if is_classification_metatype(metatype, strict=True)}
        self.columns = y.columns

        for col, s in y.items():
            if col in self.label_encoders:
                metatype, le = self.label_encoders[col]
                cat = s.dropna().unique()
                if metatype == Metatype.BINARY and len(cat) != 2:
                    raise ValueError('binary feature must have 2 categories: %s' % col)
                le.fit(cat)

        return self

    def transform(self, y):
        """
        Label encodes 'binary' and 'categorical' features, while 'numerical' features are represented as floats. 

        Args:
            y: DataFrame with same number of columns as metatypes.

        Returns:
            DataFrame.
        """
        if not isinstance(y, pd.DataFrame):
            raise TypeError('y must be a dataframe')
        if y.shape[1] != len(self.y_metatypes):
            raise ValueError('wrong number of columns')

        processed = []
        for col, s in y.items():
            if col in self.label_encoders:
                le = self.label_encoders[col][1]
                processed.append(pd.Series(TargetTransformer._label_transform(s, le), dtype=int, name = col))
            else:
                processed.append(s.astype(float))

        return pd.concat(processed, axis=1)

    def inverse_transform(self, y, index = None):
        """
        Inverse transforms 'binary' and 'categorical' features, while 'numerical' features are represented as floats.

        Args:
            y: DataFrame or numpy array with y.shape[1] == len(y_metatypes).
            index: optional index for DataFrame.

        Returns:
            DataFrame.

        Note:
            The output of this transformation should be the output of the algo.
        """
        if isinstance(y, np.ndarray):
            y = pd.DataFrame(y, index = index, columns = self.columns)
        elif list(y.columns) != list(self.columns):
            raise ValueError('unexpected columns in y')

        processed = []
        for col, s in y.items():
            if col in self.label_encoders:
                le = self.label_encoders[col][1]
                processed.append(pd.Series(TargetTransformer._label_inverse_transform(s, le), index = y.index, dtype = 'category', name = col))
            else:
                processed.append(s.astype(float))

        return pd.concat(processed, axis=1)

    def classes(self):
        """
        Returns classes\_ for underlying LabelEncoders in sequence of the (classification-metatype) columns in y.

        Returns:
            List of (col, classes\_) tuples.

        Note:
            This is useful for constructing predict_proba and decision_function outputs, where classes\_ should be used as columns in a DataFrame.
        """
        return [(col, le[1].classes_) for col, le in self.label_encoders.items()]

def group_metatypes(columns, metatypes):
    """Group columns according to metatypes.

    Args:
        columns: iterable.
        metatypes: iterable with same length as columns.

    Returns: 
        dict with metatype keys and lists with columns as values
    """
    groups = dict()
    for col, metatype in zip(columns, metatypes):
        if metatype in groups:
            groups[metatype].append(col)
        else:
            groups[metatype] = [col]
    return groups

def _expand_dim(x):
    if x.ndim == 1:
        return np.expand_dims(x, 1)
    return x

def _check_metatypes(metatypes):
    """checks if metatypes are valid.

    Args:
        metatypes: iterable.
    
    Returns: None

    Raises: 
        ValueError
    """
    for x in metatypes:
        Metatype(x)


from abc import ABC, abstractmethod

import re
import time
import functools
import pandas as pd

import requests

BASE_URL = 'https://api.algosquare.ai/api'

class ApiObject(ABC):
    def __init__(self, entry):
        if type(self) == ApiObject:
            raise RuntimeError('base class should not be instantiated')

        for k, v in entry.items():
            if k.startswith('time_'):
                setattr(self, k, pd.to_datetime(v))
            else:
                setattr(self, k, v)

    def __str__(self):
        return str(self.__dict__)

    def update(self, entry):
        for k, v in entry.items():
            if k.startswith('time_'):
                setattr(self, k, pd.to_datetime(v))
            else:
                setattr(self, k, v)

class AuthenticationError(Exception):
    pass

class ApiError(Exception):
    pass

def check_api_call(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from .. import API_KEY

        if API_KEY is None:
            raise ValueError('API_KEY is not set')
        if not isinstance(API_KEY, str):
            raise TypeError('API_KEY must be a string')
        if len(API_KEY) != 64:
            raise ValueError('API_KEY must be 64 characters')

        r = func(API_KEY, *args, **kwargs)
        
        if r.status_code == 403:
            raise AuthenticationError(r.text)
        if not r.ok:
            raise ApiError(r.text)

        return r.json()
    return wrapper

def urljoin(*args):
    return '/'.join(args)

def fullurl(url):
    return urljoin(BASE_URL, url)

@check_api_call
def api_get(api_key, url):
    return requests.get(fullurl(url), headers=dict(Authorization = api_key))

@check_api_call
def api_post(api_key, url, **kwargs):
    return requests.post(fullurl(url), headers=dict(Authorization = api_key), **kwargs)

@check_api_call
def api_put(api_key, url, **kwargs):
    return requests.put(fullurl(url), headers=dict(Authorization = api_key), **kwargs)

@check_api_call
def api_delete(api_key, url):
    return requests.delete(fullurl(url), headers=dict(Authorization = api_key))



"""Metrics to evaluate models."""

import numpy as np

from sklearn.metrics._scorer import SCORERS, get_scorer, _PredictScorer, _ProbaScorer, _ThresholdScorer
from sklearn.utils.multiclass import type_of_target 

from ..base.metric import Metric
from ..base.types import Metatype, is_target_metatype

class SklearnMetric(Metric):
    def __init__(self, scorer_name, metatypes, aggregation = 'mean'):
        """
        Invokes underlying score function.

        Args:
            scorer_name: sklearn scorer.
            metatypes: Metatype or set thereof.
            aggregation: combine multi-output problems using 'min','max','mean','median' or None.
        """
        scorer = get_scorer(scorer_name)
        if isinstance(scorer, _ProbaScorer):
            prediction_method = 'predict_proba'
        elif isinstance(scorer, _ThresholdScorer):
            prediction_method = 'decision_function'
        else:
            prediction_method = 'predict'

        if aggregation not in ('min','max','mean','median', None):
            raise ValueError('aggregation must be min, max, mean, median or None')

        super().__init__(_map_scorer_name(scorer_name), metatypes, prediction_method, scorer._sign == 1)
        self._aggregation = aggregation
        self._scorer_name = scorer_name
        self._score_func = scorer._score_func
        self._kwargs = scorer._kwargs

        self.pos_label = None

    def to_dict(self):
        output = super().to_dict()
        if self.pos_label is not None:
            output['kwargs'] = dict(pos_label = self.pos_label)
        return output

    def score(self, y, pred):
        """
        Invokes underlying score function.

        Args:
            y: targets.
            pred: model predictions.

        Returns:
            float.
        """
        if self._metatypes == {Metatype.BINARY} and self.pos_label is None:
            raise RuntimeError('pos_label missing')

        if y.shape[1] == 1:
            return self._score(y, pred, self.pos_label)

        if self._prediction_method == 'predict':
            scores = [self._score(y[col], pred[col], self._get_pos_label(i)) for i, col in enumerate(y.columns)]
        else:
            scores = [self._score(y[col], pred[i], self._get_pos_label(i)) for i, col in enumerate(y.columns)]

        if self._aggregation is None:
            return scores

        return getattr(np, self._aggregation)(scores)

    def _score(self, y, pred, pos_label = None):
        if pos_label is None:
            return self._score_func(y, pred, **self._kwargs)

        if self._prediction_method == 'predict':
            return self._score_func(y == pos_label, pred == pos_label, **self._kwargs)

        return self._score_func(y == pos_label, pred[pos_label], **self._kwargs)

    def _get_pos_label(self, i):
        if self.pos_label is None:
            return None

        return self.pos_label[i]

def _map_scorer_name(name):
    return name.replace('neg_', '')

REGRESSION_SCORERS = {
 'explained_variance',
 'max_error',
 'neg_mean_absolute_error',
 'neg_mean_absolute_percentage_error',
 'neg_mean_gamma_deviance',
 'neg_mean_poisson_deviance',
 'neg_mean_squared_error',
 'neg_mean_squared_log_error',
 'neg_median_absolute_error',
 'neg_root_mean_squared_error',
 'r2'}

CLUSTERING_SCORERS = {
 'adjusted_mutual_info_score',
 'adjusted_rand_score',
 'completeness_score',
 'fowlkes_mallows_score',
 'homogeneity_score',
 'mutual_info_score',
 'normalized_mutual_info_score',
 'rand_score',
 'v_measure_score'}

CLASSIFICATION_SCORERS = {
 'accuracy',
 'average_precision',
 'balanced_accuracy',
 'f1',
 'f1_macro',
 'f1_micro',
 'f1_samples',
 'f1_weighted',
 'jaccard',
 'jaccard_macro',
 'jaccard_micro',
 'jaccard_samples',
 'jaccard_weighted',
 'neg_brier_score',
 'neg_log_loss',
 'precision',
 'precision_macro',
 'precision_micro',
 'precision_samples',
 'precision_weighted',
 'recall',
 'recall_macro',
 'recall_micro',
 'recall_samples',
 'recall_weighted',
 'roc_auc',
 'roc_auc_ovo',
 'roc_auc_ovo_weighted',
 'roc_auc_ovr',
 'roc_auc_ovr_weighted',
 'top_k_accuracy'}

#exclusively 2 classes
XBINARY_SCORERS = {
 'average_precision',
 'f1',
 'jaccard',
 'neg_brier_score',
 'precision',
 'recall',
 'roc_auc',
}

#exclusively more than 2 classes
XCATEGORICAL_SCORERS = {
 'f1_macro',
 'f1_micro',
 'f1_weighted',
 'jaccard_macro',
 'jaccard_micro',
 'jaccard_weighted',
 'precision_macro',
 'precision_micro',
 'precision_weighted',
 'recall_macro',
 'recall_micro',
 'recall_weighted',
 'roc_auc_ovo',
 'roc_auc_ovo_weighted',
 'roc_auc_ovr',
 'roc_auc_ovr_weighted',
 'top_k_accuracy'}

EXCLUDE_SCORERS = {
 'f1_samples',
 'jaccard_samples',
 'precision_samples',
 'recall_samples',
 'top_k_accuracy'}

METRICS = dict()

for scorer_name in SCORERS:
    if scorer_name not in EXCLUDE_SCORERS:
        if scorer_name in REGRESSION_SCORERS:
            metric = SklearnMetric(scorer_name, Metatype.NUMERICAL)
            METRICS[str(metric)] = metric
        elif scorer_name in CLASSIFICATION_SCORERS:
            if scorer_name in XBINARY_SCORERS:
                metric = SklearnMetric(scorer_name, {Metatype.BINARY})
            elif scorer_name in XCATEGORICAL_SCORERS:
                metric = SklearnMetric(scorer_name, {Metatype.CATEGORICAL})
            else:
                metric = SklearnMetric(scorer_name, {Metatype.BINARY, Metatype.CATEGORICAL})
            METRICS[str(metric)] = metric

def get_metric(metric):
    """
    Returns metric from string.

    Args:
        metric: string

    Returns:
        Metric.

    Raises: 
        KeyError.
    """
    try:
        return METRICS[metric]
    except KeyError:
        raise KeyError(f'valid metrics are: {sorted(list(METRICS))}')

def get_metatype_metrics(metatype):
    """
    Returns metrics that support a given metatype.

    Args:
        metatype: Metatype

    Returns:
        Metric.

    Raises: 
        KeyError.
    """
    if not is_target_metatype(metatype, strict = True):
        raise ValueError('invalid target metatype')

    return {name : metric for name, metric in METRICS.items() if metric.supports_metatype(metatype)}


def _expand_dim(x):
    if x.ndim == 1:
        return np.expand_dims(x, 1)
    return x



"""Test algo functionality."""
import inspect
from .errors import TestError
from .tabular import test_tabular_algo
from ..base.algo import Algo
from ..base.tabular import TabularClassificationAutoML, TabularRegressionAutoML

def test_algo(algo, output_dir, timeout = 300, nan_fraction = 0.1):
    """
    Function for testing algo functionality.

    Args:
        algo: Algo class.
        output_dir: local directory for saving model.
        timeout: search time in seconds.
        nan_fraction: probability of null values in data.

    Returns:
        None.
    """
    if issubclass(algo, TabularClassificationAutoML) or issubclass(algo, TabularRegressionAutoML):
        return test_tabular_algo(algo, output_dir, timeout, nan_fraction)
    else:
        raise TestError('invalid algo class')

def test_package(package, output_dir, **kwargs):
    """
    Function for testing algos in package.

    Args:
        package: module.
        output_dir: local directory for saving model.
        kwargs: keyword args passed to test_algo

    Returns:
        None.
    """
    for field in dir(package):
        attr = getattr(package, field)
        if inspect.isclass(attr) and issubclass(attr, Algo):
            test_algo(attr, output_dir, **kwargs)



"""Types and auxiliary functions."""

from enum import IntEnum

class Metatype(IntEnum):
    """Metatype of data."""
    VOID = 0
    NONE = 1
    CONSTANT = 2
    NUMERICAL = 3
    CATEGORICAL = 4
    BINARY = 5
    DATETIME = 6
    TIMESTAMP = 7
    DELIM_PIPE = 8
    DELIM_SEMICOLON = 9
    LIST = 10
    TEXT = 11

DELIMITER_METATYPES = {Metatype.DELIM_PIPE, Metatype.DELIM_SEMICOLON}
TARGET_METATYPES = {Metatype.BINARY, Metatype.CATEGORICAL, Metatype.NUMERICAL}

def is_delimiter_metatype(metatype, strict = False):
    """Checks if metatype is delimiter metatype.

    Args:
        metatype: Metatype or int.
        strict: checks if metatype is Metatype.

    Returns:
        bool.
    """
    if strict and not isinstance(metatype, Metatype):
        return False

    try:
        return Metatype(metatype) in DELIMITER_METATYPES
    except ValueError:
        return False

def get_delimiter(metatype):
    """Maps delimiter metatype to delimiter.

    Args:
        metatype: Metatype or int.

    Returns:
        str.

    Raises:
        ValueError
    """
    metatype = Metatype(metatype)
    if metatype == Metatype.DELIM_PIPE:
        return '|'

    if metatype == Metatype.DELIM_SEMICOLON:
        return ';'

    raise ValueError('invalid delimiter metatype')

def is_classification_metatype(metatype, strict = False):
    """Checks if metatype is valid classification target metatype.

    Args:
        metatype: Metatype or int.
        strict: checks if metatype is Metatype.

    Returns:
        bool.
    """
    if strict and not isinstance(metatype, Metatype):
        return False

    try:
        return Metatype(metatype) in (Metatype.BINARY, Metatype.CATEGORICAL)
    except ValueError:
        return False

def is_target_metatype(metatype, strict = False):
    """Checks if metatype is valid target metatype.

    Args:
        metatype: Metatype or int.
        strict: checks if metatype is Metatype.

    Returns:
        bool.
    """
    if strict and not isinstance(metatype, Metatype):
        return False

    try:
        return Metatype(metatype) in TARGET_METATYPES
    except ValueError:
        return False

def string_to_metatype(string):
    """Converts string to Metatype.

    Args:
        string: str.

    Returns:
        Metatype.

    Raises:
        TypeError, ValueError
    """
    if not isinstance(string, str):
        raise TypeError('string must be a str')

    return Metatype[string.upper()]

class PredictionMethod(IntEnum):
    """Prediction method of predictor."""
    PREDICT = 0
    PREDICT_PROBA = 1
    DECISION_FUNCTION = 2

def is_prediction_method(prediction_method, strict = False):
    """Checks if prediction_method is valid.

    Args:
        prediction_method: PredictionMethod or int.
        strict: checks if prediction_method is PredictionMethod.

    Returns:
        bool.
    """
    if strict and not isinstance(prediction_method, PredictionMethod):
        return False

    try:
        PredictionMethod(prediction_method)
    except ValueError:
        return False

    return True



import os
import io
import requests

from .api import api_put

def upload_dataframe(namespace, df):
    with io.BytesIO() as f:
        df.to_csv(f, index=False)
        response = api_put('api/scratchpad', json=dict(filename = f'{namespace}.csv'))
        requests.put(response['url'], data=f.getvalue())
        return response['key']

def upload_file(filepath):
    with open(filepath, 'rb') as f:
        response = api_put('api/scratchpad', json=dict(filename = os.path.split(filepath)[1]))
        requests.put(response['url'], data=f)
        return response['key']


import numpy as np
import pandas as pd
from sklearn.utils import check_random_state

def calculate_metric_stats(y, pred, metric, num_samples = None, num_repeats = 1000):
    """
    Calculate metric stats.

    Args:
        y: targets.
        pred: output of Predictor.
        metric: Metric
        num_samples: number of samples to sample.
        num_repeats: number of experiments.

    Returns:
        dict.

    Raises:
        ValueError
    """
    y_np = np.array(y)
    pred_np = np.array(pred)

    random_state = check_random_state(None)

    if num_samples is None:
        num_samples = y.shape[0]
    else:
        if num_samples > y.shape[0]:
            raise ValueError('num_samples greater than samples in y')

    scores = np.zeros(num_repeats)
    idx = np.arange(y.shape[0])
    for n_round in range(num_repeats):
        choice_idx = random_state.choice(idx, num_samples)

        y_sample = y_np[choice_idx]
        pred_sample = pred_np[choice_idx]

        scores[n_round] = metric.score(pd.DataFrame(y_sample, columns = y.columns), pd.DataFrame(pred_sample, columns = pred.columns))

    output = dict()
    output['num_samples'] = num_samples
    output['num_repeats'] = num_repeats
    output['mean'] = np.mean(scores)
    output['std'] = np.std(scores)
    return output


"""AlgoSquare Predictor API."""
import os
import io
import json
import uuid
import pandas as pd

from ..base.types import Metatype
from ..base.tabular import TabularClassifier, TabularRegressor
from ..base.metric import Metric
from ..metrics import get_metric

from .common import upload_file
from .api import ApiObject, api_post, api_get, api_delete
from .model import Model

def get_predictors():
    """
    Gets all predictors.
    
    Returns:
        List of Predictors.
    """
    return [Predictor(x) for x in api_get('api/predictors')]

def add_model(predictor, output_dir, tag = None, metric = None, metric_stats = None, predictor_id = None):
    """
    Add model to latest or specific Predictor.
    
    Args:
        predictor: TabularClassifier or TabularRegressor.
        output_dir: path for temp files.   
        tag: model tag.
        metric: Metric.
        metric_stats: output from tools.analytics.calculate_metric_stats.
        predictor_id: string.

    Returns:
        Model.

    Raises:
        TypeError, ValueError
    """
    if not _is_valid_predictor(predictor):
        raise TypeError('invalid predictor')  

    if predictor_id is not None and not isinstance(predictor_id, str):
        raise TypeError('invalid predictor_id')

    settings = dict()
    if tag is not None:
        if len(str(tag)) > 64:
            raise ValueError('tag must not be longer than 64')
        settings['tag'] = str(tag)

    if metric is not None:
        if not isinstance(metric, Metric):
            raise TypeError('metric should be a Metric')

        metric_dict = metric.to_dict()

        if metric_stats is not None:
            metric_dict['stats'] = metric_stats

        settings['metric'] = metric_dict

    filename = uuid.uuid4().hex + '.mdl'
    model_path = os.path.join(output_dir, filename)
    predictor.save(model_path)

    payload = dict(settings = settings, class_name = type(predictor).__name__)
    payload['file'] = dict(filename = filename, key = upload_file(model_path))
    if predictor_id is not None:
        payload['predictor_id'] = predictor_id

    os.remove(model_path)
    return Model.load(api_post('api/predictors', json=payload))

class Predictor(ApiObject):
    def add_model(predictor, output_dir, tag = None, metric = None, metric_stats = None):
        """
        Add model to specific Predictor.
        
        Args:
            predictor: TabularClassifier or TabularRegressor.
            output_dir: path for temp files.   
            tag: model tag.
            metric: Metric.
            metric_stats: output from tools.analytics.calculate_metric_stats.

        Returns:
            Model.

        Raises:
            TypeError, ValueError
        """
        return add_model(predictor, output_dir, tag, metric, metric_stats, self.predictor_id)

    @classmethod
    def get(cls, predictor_id):
        """
        Gets specific predictor.
        
        Args:
            predictor_id: string.

        Returns:
            Predictor.
        """
        return cls(api_get(f'api/predictors/{predictor_id}'))

    def get_models(self):
        """
        Gets models for predictor.
        
        Returns:
            List of Models.
        """
        return [Model.load(x) for x in api_get(f'api/predictors/{self.predictor_id}/models')]

def _is_classifier(predictor):
    for base_class in [TabularClassifier]:
        if isinstance(predictor, base_class):
            return True
    return False

def _is_valid_predictor(predictor):
    for base_class in [TabularClassifier, TabularRegressor]:
        if isinstance(predictor, base_class):
            return True
    return False


"""AlgoSquare Task API."""
from .api import ApiObject, api_get, api_post, api_delete
from .model import Model

class Task(ApiObject):
    """Task class."""
    @classmethod
    def get(cls, task_id):
        """
        Gets specific task.
        
        Args:
            task_id: string.

        Returns:
            Task.
        """
        return cls(api_get(f'api/tasks/{task_id}'))

    def refresh(self):
        """reloads task."""
        self.update(api_get(f'api/tasks/{self.task_id}'))

    def start_automl(self, run_time = 60, max_algos = 10, server_cost = 0.2, max_algo_fee = 5):
        """
        Builds model for task.
        
        Args:
            run_time: time in minutes.
            max_algos: maximum number of algos to use.
            server_cost: maximum server cost per hour in USD.
            max_algo_fee: maximum hourly rate of deployment.

        Raises:
            RuntimeError, ValueError.
        """
        if self.status != 'inactive':
            raise RuntimeError('status must be inactive')

        run_time = int(run_time)
        if run_time < 5:
            raise ValueError('run_time must be minimum 5')

        max_algos = int(max_algos)
        if max_algos < 1:
            raise ValueError('max_algos must be minimum 1')

        server_cost = float(server_cost)
        if server_cost < 0.2:
            raise ValueError('server_cost must be a minimum of 0.2')

        max_algo_fee = float(max_algo_fee)
        if max_algo_fee < 0:
            raise ValueError('max_algo_fee must not be negative')

        self.update(api_post(f'api/tasks/{self.task_id}/automl', json=dict(run_time = run_time, max_algos = max_algos, server_cost = server_cost, max_algo_fee = max_algo_fee)))

    def stop_automl(self):
        """
        Stops model building.
        
        Raises:
            RuntimeError.
        """
        if self.status != 'inactive':
            raise RuntimeError('status is inactive')

        self.update(api_delete(f'api/tasks/{self.task_id}/automl'))

    def get_models(self):
        """
        Gets models for task.
        
        Returns:
            List of Models.
        """
        return [Model.load(x) for x in api_get(f'api/tasks/{self.task_id}/models')]